/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
hendrik.lipka@gmx.de
*/

package de.hendriklipka.ldse.server.dataserver;

import java.sql.*;
import java.util.*;
import java.io.IOException;
import java.io.PrintWriter;
import org.w3c.dom.*;
import de.hendriklipka.ldse.common.*;
import de.hendriklipka.ldse.server.data.document.*;
import de.hendriklipka.ldse.server.data.server.*;

/**
 * Klasse, die die Verbindung zur Datenbank abkapselt
 */
public class JDataStoreConnection implements DBConnection
{
    protected Connection            _con=null;
    protected boolean               _useJoinedQueries=false;

    // (fast) alle SQL-Statements in vorkompilierter Form
    protected PreparedStatement     _getDocumentForID=null;
    protected PreparedStatement     _getDocumentForURL=null;
    protected PreparedStatement     _getDocumentForSpider=null;
    protected PreparedStatement     _getDocumentIDForTriplet=null;
    protected PreparedStatement     _getTripletCount=null;
    protected PreparedStatement     _getTripletCountForID=null;
    protected PreparedStatement     _getTripletsForDocumentID=null;
    protected PreparedStatement     _updateDocument=null;
    protected PreparedStatement     _insertDocument=null;
    protected PreparedStatement     _deleteDocumentForID=null;
    protected PreparedStatement     _deleteDocumentForName=null;
    protected PreparedStatement     _markDocuments=null;
    protected PreparedStatement     _unmarkDocument=null;
    protected PreparedStatement     _sweepDocuments=null;
    protected PreparedStatement     _getDocumentsToSweep=null;
    protected PreparedStatement     _updateTriplet=null;
    protected PreparedStatement     _insertTriplet=null;
    protected PreparedStatement     _deleteTriplet=null;
    protected PreparedStatement     _tripletsFiles=null;
    protected PreparedStatement     _sweepTriplets=null;
    protected PreparedStatement     _getServerIDForTriplet=null;
    protected PreparedStatement     _getSubServers=null;
    protected PreparedStatement     _getSubServerCount=null;
    protected PreparedStatement     _getSubServerData=null;
    protected PreparedStatement     _updateServer=null;
    protected PreparedStatement     _insertServer=null;
    protected PreparedStatement     _deleteServerTriplets=null;
    protected PreparedStatement     _insertServerTriplets=null;
    protected PreparedStatement     _createDocCount=null;
    protected PreparedStatement     _createWordCount=null;
    protected PreparedStatement     _createTripletCount=null;
    protected PreparedStatement     _getTripletsForServer=null;
    protected PreparedStatement     _createTripletData=null;
    protected PreparedStatement     _deleteSubServer=null;
    protected PreparedStatement     _deleteSubServerTriplets=null;
    protected PreparedStatement     _getSubServerIDByName=null;
    protected PreparedStatement     _tripletsServers=null;
    protected PreparedStatement     _getWordTriplet=null;
    protected PreparedStatement     _insertWordTriplet=null;
    protected PreparedStatement     _deleteWordTriplets=null;
    protected PreparedStatement     _getWordForID=null;
    protected PreparedStatement     _getWordsForTripletID=null;
    protected PreparedStatement     _deleteWord=null;
    protected PreparedStatement     _insertWord=null;
    protected PreparedStatement     _getWordForName=null;
    protected PreparedStatement     _getWordDocument=null;
    protected PreparedStatement     _insertWordDocument=null;
    protected PreparedStatement     _updateWordDocument=null;
    protected PreparedStatement     _deleteWordDocuments=null;
    
    protected int _documentID=-1;
    protected int _wordID=-1;

    // Objekte fuer das Singleton
    protected static JDataStoreConnection   _engine=null;
    protected static Object         _syncObj=new Object();

    // Konfigurationsdaten
    String                          driver="";
    String                          url="";
    String                          user=null;
    String                          passwd=null;

    // Variablen fuer Locking
    // Lock fr die Dokumente
    ReaderWriterLock                _lock=new ReaderWriterLock();

    // Lock fr die SubServer
    ReaderWriterLock                _lockSub=new ReaderWriterLock();

    // Statistik
    protected long                  _calls=0;
    protected long                  _read=0;
    protected long                  _write=0;
    protected long                  _time=0;
    protected long                  _readTime=0;
    protected long                  _writeTime=0;

    /**
     * Constructor declaration
     *
     *
     * @param configNode
     *
     * @see
     */
    protected JDataStoreConnection(Node configNode) throws Exception
    {
        try
        {

            // Konfiguration lesen
            NamedNodeMap    nnl=configNode.getAttributes();
            Node            attr=nnl.getNamedItem("usejoinedqueries");

            if (null!=attr)
            {
                String  att=attr.getNodeValue();

                if (att.equals("yes"))
                {
                    _useJoinedQueries=true;
                }
                else if (att.equals("no"))
                {
                    _useJoinedQueries=false;
                }
            }

            NodeList    nl=configNode.getChildNodes();

            for (int i=0; i<nl.getLength(); i++)
            {
                Node    node=nl.item(i);

                if (node.getNodeType()!=node.ELEMENT_NODE)
                {
                    continue;
                }

                String  nn=node.getNodeName();

                if (nn.equals("driverclass"))
                {
                    driver=de.hendriklipka.ldse.common.Util.getTextData(node);
                }
                else if (nn.equals("databaseurl"))
                {
                    url=de.hendriklipka.ldse.common.Util.getTextData(node);
                }
                else if (nn.equals("username"))
                {
                    user=de.hendriklipka.ldse.common.Util.getTextData(node);
                }
                else if (nn.equals("password"))
                {
                    passwd=de.hendriklipka.ldse.common.Util.getTextData(node);
                }
            }

            initDataBase(driver,url,user,passwd);

            // Statements vorbereiten
            _getTripletCountForID=
                _con.prepareStatement("SELECT NUM FROM TRIPLETDOCS WHERE ID_TRIPLET=? AND ID_FILE=?");
            _insertDocument=
                _con.prepareStatement("INSERT INTO DOCUMENTS VALUES(?,?,?,?,?,?,?,?,?,?,?,?,0)");
            _getDocumentForID=
                _con.prepareStatement("SELECT * FROM DOCUMENTS WHERE DOC_ID=?");
            _getDocumentForURL=
                _con.prepareStatement("SELECT * FROM DOCUMENTS WHERE URLNAME=?");
            _getDocumentForSpider=
                _con.prepareStatement("SELECT * FROM DOCUMENTS WHERE SPIDERID=?");
            _getDocumentIDForTriplet=
                _con.prepareStatement("SELECT ID_FILE,NUM FROM TRIPLETDOCS WHERE ID_TRIPLET=?");
            _getTripletsForDocumentID=
                _con.prepareStatement("SELECT ID_TRIPLET,NUM FROM TRIPLETDOCS WHERE ID_FILE=?");
            _getTripletCount=
                _con.prepareStatement("SELECT SUM(NUM) FROM TRIPLETDOCS WHERE ID_TRIPLET=?");
            _updateDocument=
                _con.prepareStatement("UPDATE DOCUMENTS SET TITLE=?,EXCERPT=?,DOC_SIZE=?,CHECKSUM_LOW=?,CHECKSUM_HIGH=?,LASTCHANGE=?,MIME_TYPE=?,TRIPLETCOUNT=?,WORDCOUNT=?,DOC_MARK=0 WHERE DOC_ID=?");
            _deleteDocumentForID=
                _con.prepareStatement("DELETE FROM DOCUMENTS WHERE DOC_ID=?");
            _deleteDocumentForName=
                _con.prepareStatement("DELETE FROM DOCUMENTS WHERE URLNAME=?");
            _markDocuments=
                _con.prepareStatement("UPDATE DOCUMENTS SET DOC_MARK=1 WHERE SPIDERID=?");
            _unmarkDocument=
                _con.prepareStatement("UPDATE DOCUMENTS SET DOC_MARK=0 WHERE URLNAME=?");
            _getDocumentsToSweep=
                _con.prepareStatement("SELECT DOC_ID FROM DOCUMENTS WHERE DOC_MARK=1 AND SPIDERID=?");
            _sweepDocuments=
                _con.prepareStatement("DELETE FROM DOCUMENTS WHERE DOC_MARK=1 AND SPIDERID=?");
            _insertTriplet=
                _con.prepareStatement("INSERT INTO TRIPLETDOCS VALUES(?,?,?)");
            _updateTriplet=
                _con.prepareStatement("UPDATE TRIPLETDOCS SET NUM=? WHERE ID_TRIPLET=? AND ID_FILE=?");
            _deleteTriplet=
                _con.prepareStatement("DELETE FROM TRIPLETDOCS WHERE ID_FILE=?");
//            _tripletsFiles=
//                _con.prepareStatement("SELECT DISTINCT ID_FILE FROM TRIPLETDOCS WHERE NOT EXISTS (SELECT * FROM DOCUMENTS WHERE TRIPLETDOCS.ID_FILE=DOCUMENTS.DOC_ID)");
//            _tripletsServers=
//                _con.prepareStatement("SELECT DISTINCT ID_SERVER FROM SERVERTRIPLETS WHERE NOT EXISTS (SELECT * FROM QUERYSERVER WHERE SERVERTRIPLETS.ID_SERVER=QUERYSERVER.ID_SERVER)");
//            _sweepTriplets=
//                _con.prepareStatement("DELETE FROM TRIPLETDOCS WHERE ID_FILE IN (SELECT ID FROM DOCUMENTS WHERE DOC_MARK=1 AND SPIDERID=?)");
            _getServerIDForTriplet=
                _con.prepareStatement("SELECT ID_SERVER,NUM,DOCNUM,MAXNUM FROM SERVERTRIPLETS WHERE ID_TRIPLET=?");
            _getSubServers=
                _con.prepareStatement("SELECT ID_SERVER FROM QUERYSERVER");
            _getSubServerCount=
                _con.prepareStatement("SELECT COUNT(ID_SERVER) FROM QUERYSERVER");
            _getSubServerData=
                _con.prepareStatement("SELECT * FROM QUERYSERVER WHERE ID_SERVER=?");
            _updateServer=
                _con.prepareStatement("UPDATE QUERYSERVER SET NAME=?, PORT=?, DOCUMENTS=?, TRIPLETS=?"
                                      +", WORDS=?, SUBSERVERS=?, SUBDOCS=?, SUBWORDS=?, SUBTRIPLETS=? WHERE ID_SERVER=?");
            _insertServer=
                _con.prepareStatement("INSERT INTO QUERYSERVER VALUES (?,?,?,?,?,?,?,?,?,?)");
            _deleteServerTriplets=
                _con.prepareStatement("DELETE FROM SERVERTRIPLETS WHERE ID_SERVER=?");
            _insertServerTriplets=
                _con.prepareStatement("INSERT INTO SERVERTRIPLETS VALUES (?,?,?,?,?)");
            _createWordCount=
                _con.prepareStatement("SELECT SUM(WORDCOUNT) FROM DOCUMENTS");
            _createTripletCount=
                _con.prepareStatement("SELECT SUM(TRIPLETCOUNT) FROM DOCUMENTS");
            _getTripletsForServer=
                _con.prepareStatement("SELECT * FROM SERVERTRIPLETS WHERE ID_SERVER=?");
            _createTripletData=
                _con.prepareStatement("SELECT ID_TRIPLET, SUM(NUM), MAX(NUM), COUNT(NUM) FROM TRIPLETDOCS GROUP BY ID_TRIPLET");
            _createDocCount=
                _con.prepareStatement("SELECT COUNT(DOC_ID) FROM DOCUMENTS");
            _deleteSubServer=
                _con.prepareStatement("DELETE FROM QUERYSERVER WHERE ID_SERVER=?");
            _deleteSubServerTriplets=
                _con.prepareStatement("DELETE FROM SERVERTRIPLETS WHERE ID_SERVER=?");
            _getSubServerIDByName=
                _con.prepareStatement("SELECT ID_SERVER FROM QUERYSERVER WHERE NAME=?");
            _getWordTriplet=
                _con.prepareStatement("SELECT * FROM TRIPLETWORDS WHERE ID_TRIPLET=?");
//            _getWordsForTripletID=
//                _con.prepareStatement("SELECT * FROM WORDS WHERE WORDS.ID_WORD IN (SELECT TRIPLETWORDS.ID_WORD FROM TRIPLETWORDS WHERE ID_TRIPLET=?)");
            _getWordForID=
                _con.prepareStatement("SELECT * FROM WORDS WHERE ID_WORD=?");
            _getWordForName=
                _con.prepareStatement("SELECT * FROM WORDS WHERE NAME=?");
            _getWordDocument=
                _con.prepareStatement("SELECT * FROM WORDDOCUMENTS WHERE ID_WORD=?");
            _insertWordTriplet=
                _con.prepareStatement("INSERT INTO TRIPLETWORDS VALUES(?,?,?)");
            _deleteWordTriplets=
                _con.prepareStatement("DELETE FROM TRIPLETWORDS WHERE ID_WORD=?");
            _insertWord=
                _con.prepareStatement("INSERT INTO WORDS VALUES(?,?)");
            _deleteWord=
                _con.prepareStatement("DELETE FROM WORDS WHERE ID_WORD=?");
            _insertWordDocument=
                _con.prepareStatement("INSERT INTO WORDDOCUMENTS VALUES(?,?,?)");
            _updateWordDocument=
                _con.prepareStatement("UPDATE WORDDOCUMENTS SET NUM=? WHERE ID_FILE=? AND ID_WORD=?");
            _deleteWordDocuments=
                _con.prepareStatement("DELETE FROM WORDDOCUMENTS WHERE ID_FILE=?");
        }
        catch (Exception ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));
        }
    }

    /**
     * Method declaration
     *
     *
     * @param driverclass
     * @param url
     * @param user
     * @param pwd
     *
     * @see
     */
    public void initDataBase(String driverclass,String url,String user,
                             String pwd)
    {

        // Liste aller Tabellen und ihrer Definition
        String[][]          allTables=
        {
            {
                "DOCUMENTS",
                "CREATE TABLE DOCUMENTS ("
                +"DOC_ID			INT,\n"
                +"URLNAME	    VARCHAR(400),\n"
                +"SPIDERID      INT,\n"
                +"TITLE         VARCHAR(400),\n"
                +"EXCERPT	    VARCHAR(400),\n"
                +"DOC_SIZE          INT,\n"+"CHECKSUM_LOW INT,\n"
                +"CHECKSUM_HIGH INT,\n"
                +"LASTCHANGE    DATE,\n"
                +"MIME_TYPE      VARCHAR(60),\n"
                +"TRIPLETCOUNT  INT,\n"
                +"WORDCOUNT     INT,\n"
                +"DOC_MARK          SMALLINT)"
            },
            {
                "TRIPLETDOCS",
                "CREATE TABLE TRIPLETDOCS (\n"
                +"ID_TRIPLET INT,\n"
                +"ID_FILE INT,\n"
                +"NUM INT)"
            },
            {
                "TRIPLETWORDS",
                "CREATE TABLE TRIPLETWORDS (\n"
                +"ID_TRIPLET INT,\n"
                +"ID_WORD INT,\n"+"NUM INT)\n"
            },
            {
                "WORDS",
                "CREATE TABLE WORDS (\n"
                +"NAME VARCHAR(50),"
                +"ID_WORD INT)\n"
            },
            {
                "WORDDOCUMENTS",
                "CREATE TABLE WORDDOCUMENTS (\n"+"ID_WORD INT,"+"ID_FILE INT,"
                +"NUM INT)\n"
            },
            {
                "QUERYSERVER",
                "CREATE TABLE QUERYSERVER(\n"+"ID_SERVER INT,\n"+   // sinnvollerweise aus der IP-Adresse gebildet...
                "NAME VARCHAR(150),\n"+                             // Netzwerkname!!! zwingend!
                "PORT INT,\n"+"DOCUMENTS INT,\n"+                   // Dokumente auf diesem Server
                "TRIPLETS INT,\n"+                                  // Triplets auf diesem Server
                "WORDS INT,\n"+                                     // Woerter auf diesem Server
                "SUBSERVERS INT,\n"+                                // Anzahl der untergeordneten Server
                "SUBDOCS INT,\n"+                                   // Dokumente auf untergeordneten Servern
                "SUBWORDS INT,\n"+                                  // Woerter auf untergeordneten Servern
                "SUBTRIPLETS INT)"
            },                                                      // Triplets auf untergeordneten Servern
            {
                "SERVERTRIPLETS",
                "CREATE TABLE SERVERTRIPLETS(\n"+"ID_SERVER INT,\n"
                +"ID_TRIPLET INT,\n"+"NUM INT,\n"+"MAXNUM INT,"+"DOCNUM INT)"
            },
            {
                "TRIPLETWORDSINDEX",
                "CREATE INDEX TRIPLETWORDSINDEX ON TRIPLETWORDS(ID_TRIPLET)"
            },
            {
                "WORDINDEX","CREATE INDEX WORDINDEX ON WORDS(ID_WORD)"
            },
            {
                "WORDINDEX2","CREATE INDEX WORDINDEX2 ON WORDS(NAME)"
            },
            {
                "WORDDOCUMENTSINDEX",
                "CREATE INDEX WORDDOCUMENTSINDEX ON WORDDOCUMENTS(ID_WORD)"
            },
            {
                "TRIPLETDOCSINDEX",
                "CREATE INDEX TRIPLETDOCSINDEX ON TRIPLETDOCS(ID_TRIPLET)"
            },
            {
                "DOCUMENTSNAMEINDEX",
                "CREATE INDEX DOCUMENTSNAMEINDEX ON DOCUMENTS(URLNAME)"
            },
            {
                "DOCUMENTSIDINDEX",
                "CREATE INDEX DOCUMENTSIDINDEX ON DOCUMENTS(DOC_ID)"
            },
            {
                "QUERYSERVERINDEX",
                "CREATE INDEX QUERYSERVERINDEX ON QUERYSERVER(ID_SERVER)"
            },
            {
                "SERVERTRIPLETSINDEX",
                "CREATE INDEX SERVERTRIPLETSINDEX ON SERVERTRIPLETS(ID_TRIPLET)"
            }
        };
        Statement           stmt=null;
        DatabaseMetaData    dmd=null;
        ResultSet           tables=null;
        ResultSet           indexes=null;

        System.runFinalizersOnExit(true);

        try
        {
            Log.log("database","startup",1,"creating connection for "+url);

            Class   idbClass=Class.forName(driverclass);

            Log.log("database","startup",1,"driver "+driverclass+" loaded");
            if (null!=_engine)
            {
                try
                {
                    Log.log("database","startup",1,"closing old connection:");
                    _engine._con.commit();
                    _engine._con.close();

                    _engine._con=null;

                    System.gc();
                }
                catch (Exception ex)
                {
                    ex.printStackTrace(Log.getWriter("database","error",0));
                }
            }

            // mit DB verbinden
            _con=DriverManager.getConnection(url,user,pwd);

            // Liste aller Tabellen und Indizes holen
            Log.log("database","startup",1,"checking for tables:");

            stmt=_con.createStatement();
            dmd=_con.getMetaData();

            String[]    tableTypes=
            {
                "TABLE"
            };

            tables=dmd.getTables(null,null,"%",tableTypes);

            Vector  tableNames=new Vector();

            while (tables.next())
            {
                String  tableName=tables.getString(3);
                String  cat=tables.getString(1);
                String  schem=tables.getString(2);

                tableNames.addElement(tableName);
                Log.log("database","startup",1,cat+":"+schem+":"+tableName);

                // Indizes holen und zu tableNames dazu (damit sie nicht doppelt angelegt werden)
                System.out.println(tableName);
                if (null!=schem)
                    if (schem.equals("SYS"))
                        continue;
                indexes=dmd.getIndexInfo(cat,schem,tableName,false,false);

                while (indexes.next())
                {
                    String  indexName=indexes.getString(6);

                    cat=tables.getString(1);
                    schem=tables.getString(2);

                    tableNames.addElement(indexName);
                    Log.log("database","startup",2,
                            "  "+cat+":"+schem+":"+indexName);
                }
            }

            // Kontrolle, ob alle Tabellen vorhanden sind
            for (int i=0; i<allTables.length; i++)
            {
                Log.log("database","startup",1,allTables[i][0]+":");

                boolean found=false;

                for (int j=0; j<tableNames.size(); j++)
                {
                    if (allTables[i][0].equalsIgnoreCase((String)tableNames.elementAt(j)))
                    {
                        found=true;
                    }
                }

                if (!found)
                {

                    // wenn nicht, anlegen
                    Log.log("database","startup",1," creating");
                    Log.log("database","sql",2,allTables[i][1]);

        
                    int r=stmt.executeUpdate(allTables[i][1]);

                    Log.log("database","startup",4,Integer.toString(r));
                }
                else
                {
                    Log.log("database","startup",1," exists");
                }
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));
        }
        finally
        {
            try
            {
                tables.close();
            }
            catch (Exception ex) {}

            try
            {
                indexes.close();
            }
            catch (Exception ex) {}

            try
            {
                stmt.close();
            }
            catch (Exception ex) {}

            try
            {
                _con.commit();
            }
            catch (Exception ex) {}
        }
    }

    /**
     * Method declaration
     *
     *
     * @param node
     *
     * @return
     *
     * @see
     */
    public static JDataStoreConnection openConnection(Node node)
    {
        try
        {
            if (null==_engine)
            {
                synchronized (_syncObj)
                {
                    if (null==_engine)
                    {
                        _engine=new JDataStoreConnection(node);
                    }
                }
            }

            return _engine;
        }
        catch (Exception ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return null;
        }
    }

    /**
     * Method declaration
     *
     *
     * @see
     */
    protected void finalize()
    {
        try
        {
            if (null!=_con)
            {
                _con.close();
            }

            System.gc();
        }
        catch (Exception ex) {}
    }

    /**
     * berprft, ob ein Dokument mit einer gegebenen URL existiert
     */
    public boolean existDocument(String name)
    {
        _calls++;
        _read++;

        _lock.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            _getDocumentForURL.clearParameters();
            _getDocumentForURL.setString(1,name);

            rs=_getDocumentForURL.executeQuery();

            if (false==rs.next())
            {
                return false;
            }

            return true;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return false;
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lock.finishedRead();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }
    }

    /**
     * setzt die Marke fr ein gegebenes Dokument zurck
     */
    public String unmarkDocument(String name)
    {
        _calls++;
        _write++;

        _lock.requestWrite();

        long    start=System.currentTimeMillis();

        try
        {
            _unmarkDocument.clearParameters();
            _unmarkDocument.setString(1,name);
            _unmarkDocument.executeUpdate();

            return null;
        }
        catch (Exception ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            _lock.finishedWrite();
        }
    }

    /**
     * speichert bzw. erneuert ein Dokument in der Datenbank
     */
    public String putDocument(String name,DocumentDescriptor dd,
                              WordDocument[] words)
    {
        _calls++;
        _write++;

        _lock.requestWrite();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;
        ResultSet   rs2=null;

        try
        {
            _getDocumentForURL.clearParameters();
            _getDocumentForURL.setString(1,name);

            rs=_getDocumentForURL.executeQuery();

            if (false==rs.next())
            {
                Log.log("database","query",2,"insert "+name);

                // Document existiert noch nicht -> INSERT
                // nicht existierende ID erzeugen
                // *** ANNAHME: WORDDOCUMENTS korrespondiert mit DOCUMENTS
                // d.h. hier kann gefahrlos in WORDDOCUMENTS eingefuegt werden
                String  s=dd.getUrlStr();

                // ToDo: man beachte URLs mit weniger als 15 Zeichen...
//                int     id=s.substring(s.length()-15,s.length()).hashCode();

                if (-1==_documentID)
                {
                    Statement st=_con.createStatement();
                    try
                    {
                        rs2=st.executeQuery("SELECT MAX(DOC_ID) FROM DOCUMENTS");
                        if (rs2.next())
                            _documentID=rs2.getInt(1);
                        else 
                            _documentID=0;
                        rs2.close();
                    }
                    catch (SQLException ex)
                    {
                        ex.printStackTrace(Log.getWriter("database","error",0));
                        return ex.getMessage();
                    }
                    finally
                    {
                        st.close();
                    }
                }
                _documentID++;

                Log.log("database","query",3,"id="+_documentID);

                // Triplets einfuegen
                _insertTriplet.clearParameters();
                _insertTriplet.setInt(2,_documentID);

                Enumeration e=dd.getList().getTriplets();

                while (e.hasMoreElements())
                {
                    try
                    {
                        Triplet tr=(Triplet)e.nextElement();

                        Log.log("database","sql",3,
                                "INSERT INTO TRIPLETDOCS VALUES("+tr.num+","
                                +_documentID+","+tr.count+")");
                        _insertTriplet.setInt(1,tr.num);
                        _insertTriplet.setInt(3,tr.count);
                        _insertTriplet.executeUpdate();
                    }
                    catch (Exception ex)
                    {
                        Log.log("database","error",0,ex.toString());
                        ex.printStackTrace(Log.getWriter("database","error",
                                                         0));

                        return ex.toString();
                    }
                }

                _con.commit();

                // Document erst am Schluss einfuegen, damit bei einem evt. Abbruch
                // beim naechsten Durchlauf auf jeden Fall geupdated wird
                _insertDocument.clearParameters();
                _insertDocument.setInt(1,_documentID);
                _insertDocument.setString(2,dd.getUrlStr());
                _insertDocument.setInt(3,dd.getSpiderID());
                _insertDocument.setString(4,dd.getTitle());
                _insertDocument.setString(5,dd.getExcerpt());
                _insertDocument.setInt(6,dd.getSize());

                int cs_low=(int)(0xFFFFFFFFL&dd.getChecksum());
                int cs_high=(int)(dd.getChecksum()>>32);

                Log.log("database","sql",2,
                        "checksum="+dd.getChecksum()+" low="+cs_low+" high="
                        +cs_high);
                _insertDocument.setInt(7,cs_low);   // Low
                _insertDocument.setInt(8,cs_high);  // High
                _insertDocument.setDate(9,
                                        new java.sql.Date(dd.getLastChange().getTime()));
                _insertDocument.setString(10,dd.getType());
                _insertDocument.setInt(11,dd.getTripletCount());
                _insertDocument.setInt(12,dd.getWordCount());
                Log.log("database","sql",3,
                        "INSERT INTO DOCUMENTS VALUES("+_documentID+","+dd.getUrlStr()
                        +","+dd.getSpiderID()+","+dd.getTitle()+","
                        +dd.getExcerpt()+","+dd.getSize()+","+dd.getChecksum()
                        +","+new java.sql.Date(dd.getLastChange().getTime())
                        +","+dd.getType()+","+dd.getTripletCount()+","
                        +dd.getWordCount()+",0)");

                int r=_insertDocument.executeUpdate();

                // WORDDOCUMENTS bearbeiten
                for (int i=0; i<words.length; i++)
                {
                    WordDocument    wd=words[i];
                    WordData        wd2=wd._word;

                    Log.log("database","sql",2,"SELECT * FROM WORDS WHERE NAME=\""+wd2._word+"\"");
                    _getWordForName.setString(1,"\""+wd2._word+"\"");

                    rs2=_getWordForName.executeQuery();

                    int wid=-1;

                    if (rs2.next())
                    {

                        // ID holen
                        Log.log("database","sql",2,"got word");
                        wid=rs2.getInt(2);
                        rs2.close();
                    }
                    else
                    {
                        Log.log("database","sql",2,"no word");

                        // ID erzeugen
                        rs2.close();

                        if (-1==_wordID)
                        {
                            Statement st=_con.createStatement();
                            try
                            {
                                rs2=st.executeQuery("SELECT MAX(ID_WORD) FROM WORDS");
                                if (rs2.next())
                                    _wordID=rs2.getInt(1);
                                else
                                    _wordID=0;
                                rs2.close();
                            }
                            catch (SQLException ex)
                            {
                                ex.printStackTrace(Log.getWriter("database","error",0));
                                return ex.getMessage();
                            }
                            finally
                            {
                                st.close();
                            }
                        }
                        _wordID++;
                        wid=_wordID;
                        
                        _insertWord.setString(1,"\""+wd2._word+"\"");
                        _insertWord.setInt(2,wid);

                        // Log.log("database","sql",1,wd2._word+"="+wid);
                        _insertWord.executeUpdate();

                        // und jetzt noch die Triplets einfuegen...
                        e=wd2._triplets.getTriplets();

                        while (e.hasMoreElements())
                        {
                            Triplet tr=(Triplet)e.nextElement();

                            _insertWordTriplet.setInt(1,tr.num);
                            _insertWordTriplet.setInt(2,wid);
                            _insertWordTriplet.setInt(3,tr.count);
                            _insertWordTriplet.executeUpdate();
                        }
                    }

                    _insertWordDocument.setInt(1,wid);
                    _insertWordDocument.setInt(2,_documentID);
                    _insertWordDocument.setInt(3,wd._count);
                    _insertWordDocument.executeUpdate();
                }

                _con.commit();
            }
            else
            {
                Log.log("database","query",2,"update "+name);

                // UPDATE
                int id=rs.getInt("ID");

                Log.log("database","query",3,"id="+id);

                java.util.Date  d=rs.getDate("LASTCHANGE");
                int             checksumLow=rs.getInt("CHECKSUM_LOW");
                int             checksumHigh=rs.getInt("CHECKSUM_HIGH");
                long            checksum=((long)checksumHigh)<<32;
                long            cs=(long)checksumLow;

                cs=cs&(0xFFFFFFFFL);
                checksum=checksum+cs;

                Log.log("database","sql",2,
                        "checksum="+checksum+" low="+checksumLow+" high="
                        +checksumHigh);

                int size=rs.getInt("SIZE");

                // Check, ob Abgleich noetig ist
                if ((checksum==dd.getChecksum())&&(size==dd.getSize())
                        &&(d.equals(dd.getLastChange())))
                {
                    return null;
                }

                // Triplets update
                // zuerst alle loeschen, dann neu einfuegen
                // kein Update, weil dann geloeschte Triplets nicht geloescht werden
                _deleteTriplet.clearParameters();
                _deleteTriplet.setInt(1,id);
                _deleteTriplet.executeUpdate();
                _insertTriplet.clearParameters();
                _insertTriplet.setInt(2,id);

                Enumeration e=dd.getList().getTriplets();

                while (e.hasMoreElements())
                {
                    Triplet tr=(Triplet)e.nextElement();

                    _insertTriplet.setInt(1,tr.num);
                    _insertTriplet.setInt(3,tr.count);
                    _insertTriplet.executeUpdate();
                }

                _con.commit();

                // Document erst am Schluss updaten, damit bei einem evt. Abbruch
                // beim naechsten Durchlauf auf jeden Fall geupdated wird
                _updateDocument.clearParameters();
                _updateDocument.setString(1,dd.getTitle());
                _updateDocument.setString(2,dd.getExcerpt());
                _updateDocument.setInt(3,dd.getSize());

                int cs_low=(int)(0xFFFFFFFFL&dd.getChecksum());
                int cs_high=(int)(dd.getChecksum()>>32);

                Log.log("database","sql",2,
                        "checksum="+checksum+" low="+cs_low+" high="+cs_high);
                _updateDocument.setInt(4,cs_low);   // Low
                _updateDocument.setInt(5,cs_high);  // High
                _updateDocument.setDate(6,
                                        new java.sql.Date(dd.getLastChange().getTime()));
                _updateDocument.setString(7,dd.getType());
                _updateDocument.setInt(8,dd.getTripletCount());
                _updateDocument.setInt(9,dd.getWordCount());
                _updateDocument.setInt(10,id);

                int r=_updateDocument.executeUpdate();

                _con.commit();
            }

            return null;
        }
        catch (Exception ex)
        {
            Log.log("database","error",0,ex.getMessage());
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            _lock.finishedWrite();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex)
            {
                ex.printStackTrace(Log.getWriter("database","error",0));
            }
        }
    }

    /**
     * liefert die Dokumentenbeschreibung fr ein existierendes Dokument
     * anhand der URL
     */
    public DocumentDescriptor getDocumentDescriptor(String name)
    {
        _calls++;
        _read++;

        _lock.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            _getDocumentForURL.clearParameters();
            _getDocumentForURL.setString(1,name);

            rs=_getDocumentForURL.executeQuery();

            if (false==rs.next())
            {
                return null;
            }

            int     id=rs.getInt("ID");
            int     spiderID=rs.getInt("SPIDERID");
            String  url=rs.getString("URLNAME");
            String  title=rs.getString("TITLE");

            if (null==title)
            {
                title="";
            }

            String  type=rs.getString("MIMETYPE");

            if (null==type)
            {
                type="text/plain";
            }

            String  excerpt=rs.getString("EXCERPT");

            if (null==excerpt)
            {
                excerpt="";
            }

            java.util.Date  d=rs.getDate("LASTCHANGE");
            int             checksumLow=rs.getInt("CHECKSUM_LOW");
            int             checksumHigh=rs.getInt("CHECKSUM_HIGH");
            long            checksum=((long)checksumHigh)<<32;
            long            cs=(long)checksumLow;

            cs=cs&(0xFFFFFFFFL);
            checksum=checksum+cs;

            Log.log("database","sql",4,
                    "get: checksum="+checksum+" low="+checksumLow+"/"+cs
                    +" high="+checksumHigh);

            int                 size=rs.getInt("SIZE");
            int                 count=rs.getInt("TRIPLETCOUNT");
            int                 wcount=rs.getInt("WORDCOUNT");
            DocumentDescriptor  dd=new DocumentDescriptor(id,null,url,
                    spiderID,title,excerpt,d,checksum,size);

            dd.setTripletCount(count);
            dd.setWordCount(wcount);
            dd.setType(type);

            return dd;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return null;
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lock.finishedRead();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }
    }

    /**
     * liefert die Dokumentenbeschreibung fr ein existierendes Dokument
     * anhand der ID
     */
    public DocumentDescriptor getDocumentDescriptor(int id)
    {
        _calls++;
        _read++;

        _lock.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            _getDocumentForID.clearParameters();
            _getDocumentForID.setInt(1,id);

            rs=_getDocumentForID.executeQuery();

            if (false==rs.next())
            {
                return null;
            }

            int     spiderID=rs.getInt("SPIDERID");
            String  url=rs.getString("URLNAME");
            String  title=rs.getString("TITLE");

            if (null==title)
            {
                title="";
            }

            String  type=rs.getString("MIME_TYPE");

            if (null==type)
            {
                type="text/plain";
            }

            String  excerpt=rs.getString("EXCERPT");

            if (null==excerpt)
            {
                excerpt="";
            }

            java.util.Date  d=rs.getDate("LASTCHANGE");
            int             checksumLow=rs.getInt("CHECKSUM_LOW");
            int             checksumHigh=rs.getInt("CHECKSUM_HIGH");
            long            checksum=((long)checksumHigh)<<32;
            long            cs=(long)checksumLow;

            cs=cs&(0xFFFFFFFFL);
            checksum=checksum+cs;

            Log.log("database","sql",4,
                    "get: checksum="+checksum+" low="+checksumLow+"/"+cs
                    +" high="+checksumHigh);

            int                 size=rs.getInt("DOC_SIZE");
            int                 count=rs.getInt("TRIPLETCOUNT");
            DocumentDescriptor  dd=new DocumentDescriptor(id,null,url,
                    spiderID,title,excerpt,d,checksum,size);

            dd.setTripletCount(count);
            dd.setType(type);

            return dd;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return null;
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lock.finishedRead();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }
    }

    /**
     * liefert die Gesamtsumme fr ein bestimmtes Triplet
     */
    public int getTripletCount(int tripletNum)
    {
        _calls++;
        _read++;

        _lock.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            _getTripletCount.clearParameters();
            _getTripletCount.setInt(1,tripletNum);

            rs=_getTripletCount.executeQuery();

            if (false==rs.next())
            {
                return 0;
            }

            int c=rs.getInt(1);

            return c;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return 0;
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lock.finishedRead();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }
    }

    /**
     * lscht ein Dokument
     */
    public String delete(String name)
    {
        _calls++;
        _write++;

        _lock.requestWrite();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            _getDocumentForURL.clearParameters();
            _getDocumentForURL.setString(1,name);

            rs=_getDocumentForURL.executeQuery();

            if (false==rs.next())
            {
                return "Dokument nicht gefunden";   // es ist ja nicht mehr da...
            }

            int id=rs.getInt("ID");

            rs.close();
            _deleteDocumentForID.setInt(1,id);
            _deleteDocumentForID.executeUpdate();
            _deleteTriplet.setInt(1,id);
            _deleteTriplet.executeUpdate();

            return null;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            _lock.finishedWrite();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }
    }

    /**
     * liefert alle Dokumente fr einen bestimmten Spider
     */
    public String[] getAllForSpider(int id)
    {
        _calls++;
        _read++;

        _lock.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            _getDocumentForSpider.clearParameters();
            _getDocumentForSpider.setInt(1,id);

            rs=_getDocumentForSpider.executeQuery();

            if (false==rs.next())
            {
                return null;
            }

            Vector  v=new Vector(10,10);
            String  s=rs.getString("URLNAME");

            v.addElement(s);

            while (rs.next())
            {
                s=rs.getString("URLNAME");

                v.addElement(s);
            }

            String[]    sa=new String[v.size()];

            for (int i=0; i<sa.length; i++)
            {
                sa[i]=(String)v.elementAt(i);
            }

            return sa;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return null;
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lock.finishedRead();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }
    }

    /**
     * markiert alle Dokumente eines bestimmten Spiders
     */
    public String markSpider(int id)
    {
        _calls++;
        _write++;

        // sicherstellen, dass kein anderer Write laeuft
        // kein Writelock, weil es Leseoperationen nicht beeintraechtigt
        // es sei denn, es gibt 2 Spider kit der gleichen ID
        _lock.requestRead();

        long    start=System.currentTimeMillis();

        try
        {
            _markDocuments.setInt(1,id);

            int r=_markDocuments.executeUpdate();

            Log.log("database","query",1,"marked: "+r);

            return null;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            _lock.finishedRead();
        }
    }

    /**
     * lscht die markierten Dokumente eines Spiders
     */
    public String sweepSpider(int id)
    {
        _calls++;
        _write++;

        _lock.requestWrite();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            int r=0;

            if (_useJoinedQueries)
            {
                _sweepTriplets.setInt(1,id);

                r=_sweepTriplets.executeUpdate();
            }
            else
            {
                _getDocumentsToSweep.setInt(1,id);

                rs=_getDocumentsToSweep.executeQuery();

                while (rs.next())
                {
                    int fid=rs.getInt(1);

                    _deleteTriplet.setInt(1,fid);

                    r+=_deleteTriplet.executeUpdate();
                }
            }

            _con.commit();
            Log.log("database","query",1,"sweeped triplets:  "+r);
            _sweepDocuments.setInt(1,id);

            r=_sweepDocuments.executeUpdate();

            _con.commit();
            Log.log("database","query",1,"sweeped documents: "+r);

            return null;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            _lock.finishedWrite();
        }
    }

    /**
     * liefert alle fr eine bestimmte Tripletmenge relevanten Dokumente zurck
     * dabei knnen auch zustzliche Tripletinformationen geliefert werden
     * Hinweis: die DocumentDescriptor sind leer, bis auf die ID
     * D.h., sie mssen spter noch gefllt werden!
     */
    public DocumentDescriptor[] queryDocuments(TripletList have,
                                               TripletList info)
    {
        _calls++;
        _read++;

        _lock.requestRead();

        long    start=System.currentTimeMillis();

        try
        {
            Hashtable   docs=new Hashtable();

            // fuer jedes Triplet in "have" eine Query, um alle Dokumente einzusammeln
            Enumeration e=have.getTriplets();

            while (e.hasMoreElements())
            {
                ResultSet   rs=null;

                try
                {
                    Triplet tr=(Triplet)e.nextElement();

                    _getDocumentIDForTriplet.setInt(1,tr.num);

                    rs=_getDocumentIDForTriplet.executeQuery();

                    int n=0;

                    while (rs.next())
                    {
                        n++;

                        int id=rs.getInt(1);
                        int num=rs.getInt(2);

                        Triplet res=new Triplet(tr.num,num);
                        Integer ID=new Integer(id);
                        Object  o=docs.get(ID);

                        if (null==o)
                        {
                            /*
                            DocumentDescriptor  ds=getDocumentDescriptor(id);
                            */
                            Properties p=new Properties();
                            p.put("id",Integer.toString(id));
                            DocumentDescriptor ds=new DocumentDescriptor(p);

                            docs.put(ID,ds);
                            ds.setList(new TripletList());
                            ds.getList().addTriplet(res);
                        }
                        else
                        {
                            DocumentDescriptor  ds=(DocumentDescriptor)o;

                            ds.getList().addTriplet(res);
                        }
                    }
                }
                catch (SQLException ex)
                {
                    ex.printStackTrace(Log.getWriter("database","error",0));
                }
                finally
                {
                    try
                    {
                        if (null!=rs)
                        {
                            rs.close();
                        }
                    }
                    catch (SQLException ex) {}
                }
            }       // while

            // Documents in Array kopieren
            int                     doccount=docs.size();
            DocumentDescriptor[]    dda=new DocumentDescriptor[doccount];
            int                     i=0;

            e=docs.keys();

            while (e.hasMoreElements())
            {
                Integer ID=(Integer)e.nextElement();

                dda[i++]=(DocumentDescriptor)docs.get(ID);
            }

            // fuer jedes gefundene Document die TripletCounts aus "info" hinzufuegen
            // wird andersrum erledigt:
            // wir gehen alle Triplets durch und schauen, ob die gefundenen Dokumente
            // oben schon aufgetaucht sind
            // ist vermutlich schneller, insbesondere, wenn wenig Dokumente da sind
            if (null!=info)
            {
                e=info.getTriplets();

                while (e.hasMoreElements())
                {
                    ResultSet   rs=null;

                    try
                    {
                        Triplet tr=(Triplet)e.nextElement();

                        _getDocumentIDForTriplet.setInt(1,tr.num);

                        rs=_getDocumentIDForTriplet.executeQuery();

                        while (rs.next())
                        {
                            int     id=rs.getInt(1);
                            Integer ID=new Integer(id);
                            Object  o=docs.get(ID);

                            if (null!=o)
                            {
                                int                 num=rs.getInt(2);
                                Triplet             res=new Triplet(tr.num,
                                                                    num);
                                DocumentDescriptor  ds=(DocumentDescriptor)o;

                                ds.getList().addTriplet(res);
                            }
                        }
                    }
                    catch (SQLException ex)
                    {
                        ex.printStackTrace(Log.getWriter("database","error",
                                                         0));
                    }
                    finally
                    {
                        try
                        {
                            if (null!=rs)
                            {
                                rs.close();
                            }
                        }
                        catch (SQLException ex) {}
                    }
                }   // while
            }

            return dda;
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lock.finishedRead();
        }
    }

    /**
     * liefert die IDs aller bekannten SubServer
     */
    public int[] getSubServers()
    {
        _calls++;
        _read++;

        _lockSub.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            rs=_getSubServerCount.executeQuery();

            rs.next();

            int count=rs.getInt(1);

            rs.close();

            int[]   ia=new int[count];

            count=0;
            rs=_getSubServers.executeQuery();

            while (rs.next())
            {
                ia[count]=rs.getInt(1);
                count++;
            }

            return ia;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lockSub.finishedRead();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }

        return null;
    }

    /**
     * speichert bzw. erneuert einen ServerDescriptor
     */
    public String putServerDescriptor(ServerDescriptor sd)
    {
        _calls++;
        _write++;

        Log.log("database","query",2,"putServerDescriptor");

        int id=sd.getID();

        _lockSub.requestWrite();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            _getSubServerData.setInt(1,id);

            rs=_getSubServerData.executeQuery();

            if (rs.next())
            {

                // Update
                Log.log("database","query",2,"putServerDescriptor - update");
                _updateServer.setString(1,sd.getName());
                _updateServer.setInt(2,sd.getPort());
                _updateServer.setInt(3,sd.getDocuments());
                _updateServer.setInt(4,sd.getTriplets());
                _updateServer.setInt(5,sd.getWords());
                _updateServer.setInt(6,sd.getSubServers());
                _updateServer.setInt(7,sd.getSubDocuments());
                _updateServer.setInt(8,sd.getSubWords());
                _updateServer.setInt(9,sd.getSubTriplets());
                _updateServer.setInt(10,id);
                _updateServer.executeUpdate();
                rs.close();
                Log.log("database","query",2,
                        "putServerDescriptor - deleting triplets");
                _deleteServerTriplets.setInt(1,id);
                _deleteServerTriplets.executeUpdate();
            }
            else
            {

                // Insert
                Log.log("database","query",2,"putServerDescriptor - insert");
                _insertServer.setInt(1,id);
                _insertServer.setString(2,sd.getName());
                _insertServer.setInt(3,sd.getPort());
                _insertServer.setInt(4,sd.getDocuments());
                _insertServer.setInt(5,sd.getTriplets());
                _insertServer.setInt(6,sd.getWords());
                _insertServer.setInt(7,sd.getSubServers());
                _insertServer.setInt(8,sd.getSubDocuments());
                _insertServer.setInt(9,sd.getSubWords());
                _insertServer.setInt(10,sd.getSubTriplets());
                _insertServer.executeUpdate();
            }

            Log.log("database","query",2,
                    "putServerDescriptor - creating TripletList");

            // Triplets werden immer neu eingefuegt, damit keine alten ueberbleiben
            Enumeration e=sd.getTripletList().getTriplets();

            // hier nehmen wir besser ein vorbereitetes Statement, damit das etwas schneller geht...
            _insertServerTriplets.setInt(1,id);

            int i=0;

            Log.log("database","query",2,
                    "putServerDescriptor - inserting triplets");

            while (e.hasMoreElements())
            {
                TripletEntry    tr=(TripletEntry)e.nextElement();

                _insertServerTriplets.setInt(2,tr.num);
                _insertServerTriplets.setInt(3,tr.count);
                _insertServerTriplets.setInt(4,tr.getMaxDocNum());
                _insertServerTriplets.setInt(5,tr.getDocCount());
                _insertServerTriplets.executeUpdate();

                i++;
            }

            Log.log("database","query",2,
                    "putServerDescriptor - "+i+" triplets");
            _con.commit();
        }
        catch (Exception ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            Log.log("database","query",2,"putServerDescriptor - finished ");
            _lockSub.finishedWrite();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }

        return null;
    }

    /**
     * liefert den ServerDescriptor fr einen bestimmten SubServer
     */
    public ServerDescriptor getServerDescriptor(int id)
    {
        _calls++;
        _read++;

        _lockSub.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            _getSubServerData.setInt(1,id);

            rs=_getSubServerData.executeQuery();

            if (rs.next())
            {
                String              name=rs.getString(2);
                int                 port=rs.getInt(3);
                int                 docs=rs.getInt(4);
                int                 triplets=rs.getInt(5);
                int                 words=rs.getInt(6);
                int                 subServers=rs.getInt(7);
                int                 subDocs=rs.getInt(8);
                int                 subWords=rs.getInt(9);
                int                 subTriplets=rs.getInt(10);
                ServerDescriptor    sd=new ServerDescriptor(id,name,port,docs,
                                                            words,triplets,
                                                            subDocs,subWords,
                                                            subTriplets,
                                                            subServers);

                rs.close();
                _getTripletsForServer.setInt(1,id);

                rs=_getTripletsForServer.executeQuery();

                while (rs.next())
                {
                    int id_tr=rs.getInt(2);
                    int num=rs.getInt(3);
                    int maxNum=rs.getInt(4);
                    int docNum=rs.getInt(5);

                    sd.addTriplet(id_tr,num,docNum,maxNum);
                }

                rs.close();

                return sd;
            }
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lockSub.finishedRead();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }

        return null;
    }

    /**
     * erzeugt einen ServerDescriptor, der nur den lokalen Server darstellt,
     * d.h. ohne SubServer
     */
    public ServerDescriptor createLocalServerDescriptor(int id,String name,
            int port)
    {
        _calls++;
        _read++;

        _lock.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            int docs=0;
            int words=0;
            int triplets=0;
            int subDocs=0;
            int subWords=0;
            int subTriplets=0;
            int subServers=0;

            rs=_createDocCount.executeQuery();

            rs.next();

            docs=rs.getInt(1);

            rs.close();

            rs=_createWordCount.executeQuery();

            rs.next();

            words=rs.getInt(1);

            rs.close();

            rs=_createTripletCount.executeQuery();

            rs.next();

            triplets=rs.getInt(1);

            rs.close();

            ServerDescriptor    sd=new ServerDescriptor(id,name,port,docs,
                                                        words,triplets,0,0,0,
                                                        0);

            if (_useJoinedQueries)
            {
                rs=_createTripletData.executeQuery();

                int j=0;

                while (rs.next())
                {
                    int i=rs.getInt(1);
                    int count=rs.getInt(2);
                    int maxNum=rs.getInt(3);
                    int docCount=rs.getInt(4);

                    if (0==(j%100))
                    {
                        System.err.println(j+":"+i);
                    }

                    j++;

                    sd.addTriplet(i,count,docCount,maxNum);
                }
            }
            else
            {

                // eine andere Variante, die u.U. schneller ist, weil sie kein GROUP BY verwendet
                PreparedStatement   pst1=
                    _con.prepareStatement("SELECT SUM(NUM), MAX(NUM), COUNT(NUM) FROM TRIPLETDOCS WHERE ID_TRIPLET=?");
                int                 sum=0;

                for (int i=Triplet._minNum; i<Triplet._maxNum; i++)
                {
                    try
                    {
                        if (0==(i%1000))
                        {
                            System.err.println(i+":"+sum);
                        }

                        pst1.setInt(1,i);

                        rs=pst1.executeQuery();

                        if (rs.next())
                        {
                            int count=rs.getInt(1);
                            int maxNum=rs.getInt(2);
                            int docCount=rs.getInt(3);

                            if (0!=count)
                            {
                                sum+=count;

                                sd.addTriplet(i,count,docCount,maxNum);
                            }
                        }

                        rs.close();
                    }
                    catch (SQLException ex)
                    {
                        ex.printStackTrace(Log.getWriter("database","error",
                                                         0));
                    }
                }
            }

            return sd;
        }
        catch (Exception ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lock.finishedRead();

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }

        return null;
    }

    /**
     * erzeugt einen Serverdescriptor, der diese Datenbank und alle
     * Subserver reprsentiert
     */
    public ServerDescriptor createServerDescriptor(int id,String name,
            int port)
    {
        _calls++;
        _read++;

        _lockSub.requestRead();

        long        start=System.currentTimeMillis();
        Statement   st=null;
        ResultSet   rs=null;

        try
        {
            int docs=0;
            int words=0;
            int triplets=0;
            int subDocs=0;
            int subWords=0;
            int subTriplets=0;
            int subServers=0;

            st=_con.createStatement();

            // lokaler Rechner
            rs=st.executeQuery("SELECT SUM(DOCUMENTS) FROM QUERYSERVER WHERE ID_SERVER=0");

            rs.next();

            docs=rs.getInt(1);

            rs.close();

            rs=st.executeQuery("SELECT SUM(WORDS) FROM QUERYSERVER WHERE ID_SERVER=0");

            rs.next();

            words=rs.getInt(1);

            rs.close();

            rs=st.executeQuery("SELECT SUM(TRIPLETS) FROM QUERYSERVER WHERE ID_SERVER=0");

            rs.next();

            triplets=rs.getInt(1);

            rs.close();

            // SubServer
            rs=st.executeQuery("SELECT COUNT(ID_SERVER) FROM QUERYSERVER");

            rs.next();

            subServers=rs.getInt(1);

            rs.close();

            rs=st.executeQuery("SELECT SUM(DOCUMENTS) FROM QUERYSERVER");

            rs.next();

            subDocs=rs.getInt(1);

            rs.close();

            rs=st.executeQuery("SELECT SUM(WORDS) FROM QUERYSERVER");

            rs.next();

            subWords=rs.getInt(1);

            rs.close();

            rs=st.executeQuery("SELECT SUM(TRIPLETS) FROM QUERYSERVER");

            rs.next();

            subTriplets=rs.getInt(1);

            rs.close();

            rs=st.executeQuery("SELECT SUM(SUBSERVERS) FROM QUERYSERVER");

            rs.next();

            subServers+=rs.getInt(1);

            rs.close();

            ServerDescriptor    sd=new ServerDescriptor(id,name,port,docs,
                                                        words,triplets,
                                                        subDocs,subWords,
                                                        subTriplets,
                                                        subServers);

            // hier ist keine Trennung zwischen lokal und Sub noetig, weil ohnehin
            // aufsummiert wird und die Daten bereits aggregiert vorliegen
            PreparedStatement   pst=
                _con.prepareStatement("SELECT SUM(NUM), MAX(MAXNUM), SUM(DOCNUM) FROM SERVERTRIPLETS WHERE ID_TRIPLET=?");
            int                 sum=0;

            for (int i=Triplet._minNum; i<Triplet._maxNum; i++)
            {
                try
                {
                    if (0==(i%1000))
                    {
                        System.err.println(i+":"+sum);
                    }

                    pst.setInt(1,i);

                    rs=pst.executeQuery();

                    if (rs.next())
                    {
                        int subCount=rs.getInt(1);
                        int subMaxNum=rs.getInt(2);
                        int subDocCount=rs.getInt(3);

                        if (0!=subCount)
                        {
                            sum+=subCount;

                            sd.addTriplet(i,subCount,subDocCount,subMaxNum);
                        }
                    }

                    rs.close();
                }
                catch (SQLException ex) {}
            }

            return sd;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lockSub.finishedRead();

            try
            {
                if (null!=st)
                {
                    st.close();
                }

                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}
        }

        return null;
    }

    /**
     * liefert alle fr eine bestimmte Tripletmenge relevanten SubServer zurck
     */
    public ServerDescriptor[] queryServers(TripletList have)
    {
        _calls++;
        _read++;

        _lockSub.requestRead();

        long    start=System.currentTimeMillis();

        try
        {
            Hashtable   servers=new Hashtable();

            // fuer jedes Triplet in "have" eine Query, um alle Dokumente einzusammeln
            Enumeration e=have.getTriplets();

            while (e.hasMoreElements())
            {
                ResultSet   rs=null;
                ResultSet   rs2=null;

                try
                {
                    Triplet tr=(Triplet)e.nextElement();

                    _getServerIDForTriplet.setInt(1,tr.num);

                    rs=_getServerIDForTriplet.executeQuery();

                    while (rs.next())
                    {
                        int     id=rs.getInt(1);
                        int     num=rs.getInt(2);
                        int     docs=rs.getInt(3);
                        int     maxDocs=rs.getInt(4);
                        Integer ID=new Integer(id);
                        Object  o=servers.get(ID);

                        if (null==o)
                        {
                            _getSubServerData.setInt(1,id);

                            rs2=_getSubServerData.executeQuery();

                            if (rs2.next())
                            {
                                String              name=rs2.getString(2);
                                int                 port=rs2.getInt(3);
                                int                 docC=rs2.getInt(4);
                                int                 triplets=rs2.getInt(5);
                                int                 words=rs2.getInt(6);
                                int                 subServers=rs2.getInt(7);
                                int                 subDocs=rs2.getInt(8);
                                int                 subWords=rs2.getInt(9);
                                int                 subTriplets=
                                    rs2.getInt(10);
                                ServerDescriptor    sd=
                                    new ServerDescriptor(id,name,port,docC,
                                                         words,triplets,
                                                         subDocs,subWords,
                                                         subTriplets,
                                                         subServers);

                                rs2.close();
                                sd.addTriplet(tr.num,num,docs,maxDocs);
                                servers.put(ID,sd);
                            }
                        }
                        else
                        {
                            ServerDescriptor    sd=(ServerDescriptor)o;

                            sd.addTriplet(tr.num,num,docs,maxDocs);
                        }
                    }
                }
                catch (SQLException ex)
                {
                    ex.printStackTrace(Log.getWriter("database","error",0));
                }
                finally
                {
                    try
                    {
                        if (null!=rs)
                        {
                            rs.close();
                        }

                        if (null!=rs2)
                        {
                            rs2.close();
                        }
                    }
                    catch (SQLException ex) {}
                }
            }   // while

            // Documents in Array kopieren
            int                 serverCount=servers.size();
            ServerDescriptor[]  sda=new ServerDescriptor[serverCount];
            int                 i=0;

            e=servers.keys();

            while (e.hasMoreElements())
            {
                Integer ID=(Integer)e.nextElement();

                sda[i++]=(ServerDescriptor)servers.get(ID);
            }

            return sda;
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _readTime+=(end-start);

            _lockSub.finishedRead();
        }
    }

    /**
     * liefert alle Worte, die mindestens eines der Triplets enthalten
     */
    public WordData[] getWordsForTriplets(TripletList list)
    {
        _calls++;
        _read++;
        _lockSub.requestRead();
        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try 
        {
            Hashtable words=new Hashtable();
            Enumeration e=list.getTriplets();
            while (e.hasMoreElements())
            {
                Triplet tr=(Triplet)e.nextElement();
                _getWordsForTripletID.setInt(1,tr.num);
                rs=_getWordsForTripletID.executeQuery();
                while (rs.next())
                {
                    String s=rs.getString(1);
                    int i=rs.getInt(2);
                    Integer I=new Integer(i);
                    WordData wd=new WordData();
                    wd._word=s;
                    wd._id=i;
                    wd._triplets=list;
                    words.put(I,wd);
                }
            }
            WordData[] wda=new WordData[words.size()];
            e=words.elements();
            int i=0;
            while (e.hasMoreElements())
            {
                wda[i]=(WordData)e.nextElement();
                i++;
            }
            return wda;
        }
        catch (Exception ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));
        }
        finally
        {
            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex) {}

            long    end=System.currentTimeMillis();
            _time+=(end-start);
            _readTime+=(end-start);
            _lockSub.finishedRead();
        }
        return null;
    }

    /**
     * Fuellt diese Datenstruktur auf, bis hin zum DocumentDescriptor
     */
    public void fillWordData(WordData wordData)
    {
        _calls++;
        _read++;
        _lockSub.requestRead();

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try 
        {
            _getWordDocument.setInt(1,wordData._id);
            rs=_getWordDocument.executeQuery();
            Vector v=new Vector(10,10);
            Hashtable docs=new Hashtable();
            while (rs.next())
            {
                WordDocument wd=new WordDocument();
                wd._word=wordData;
                wd._count=rs.getInt(3);
                int id=rs.getInt(2);
                Integer Id=new Integer(id);
                // Hmm. eigentlich sollten die Dokumente hier ja eindeutig sein...
                Object o=docs.get(Id);
                if (null!=o)
                {
                    wd._docDesc=(DocumentDescriptor)o;
                }
                else
                {
                    DocumentDescriptor dd=getDocumentDescriptor(id);
                    wd._docDesc=dd;
                    docs.put(Id,dd);
                }
                if (null==wd._docDesc._words)
                {
                    wd._docDesc._words=new Vector();
                }
                wd._docDesc._words.addElement(wd);
                v.addElement(wd);
            }
            WordDocument[] wda=new WordDocument[v.size()];
            for (int i=0;i<wda.length;i++)
            {
                wda[i]=(WordDocument)v.elementAt(i);
            }
            wordData._docs=wda;
        }
        catch (Exception ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));
        }
        finally
        {
            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (Exception ex)
            {
                ex.printStackTrace(Log.getWriter("database","error",0));
            }

            long    end=System.currentTimeMillis();
            _time+=(end-start);
            _readTime+=(end-start);
            _lockSub.finishedRead();
        }
    }

    /**
     * loescht alle Dokumente, die zu einem bestimmten Spider gehoeren
     */
    public String deleteSpider(int id)
    {
        _calls++;
        _write++;

        ResultSet   rs=null;
        long        start=System.currentTimeMillis();

        try
        {
            _getDocumentForSpider.setInt(1,id);

            rs=_getDocumentForSpider.executeQuery();

            while (rs.next())
            {
                try
                {
                    _lock.requestWrite();

                    int did=rs.getInt(1);

                    _deleteTriplet.setInt(1,did);
                    _deleteDocumentForID.setInt(1,did);
                    _deleteTriplet.executeUpdate();
                    _deleteDocumentForID.executeUpdate();
                }
                catch (SQLException ex)
                {
                    ex.printStackTrace(Log.getWriter("database","error",0));
                }
                finally
                {
                    _lock.finishedWrite();
                }
            }
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex)
            {
                ex.printStackTrace(Log.getWriter("database","error",0));
            }
        }

        return null;
    }

    /**
     * loescht alle Daten eines Subservers:
     * Triplets und seinen Eintrag in SUBSERVER
     */
    public String deleteSubServer(String name)
    {
        _calls++;
        _write++;

        _lockSub.requestRead();

        long        start=System.currentTimeMillis();
        int         id=-1;
        ResultSet   rs=null;

        // ID bstimmen
        try
        {
            _getSubServerIDByName.setString(1,name);

            rs=_getSubServerIDByName.executeQuery();

            if (rs.next())
            {
                id=rs.getInt(1);
            }
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            return ex.toString();
        }
        finally
        {
            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex)
            {
                ex.printStackTrace(Log.getWriter("database","error",0));
            }

            _lockSub.finishedRead();
        }

        // wenn wir eine ID haben, loeschen
        if (-1!=id)
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            return deleteSubServer(id);
        }

        long    end=System.currentTimeMillis();

        _time+=(end-start);
        _writeTime+=(end-start);

        return "Subserver nicht gefunden";
    }

    /**
     * loescht alle Daten eines Subservers:
     * Triplets und seinen Eintrag in SUBSERVER
     */
    public String deleteSubServer(int id)
    {
        _calls++;
        _write++;

        _lockSub.requestWrite();

        long    start=System.currentTimeMillis();

        try
        {
            Log.log("database","query",2,"deleteSubServer: "+id);
            _deleteSubServer.setInt(1,id);
            _deleteSubServerTriplets.setInt(1,id);
            _deleteSubServer.executeUpdate();
            _deleteSubServerTriplets.executeUpdate();
            _con.commit();

            return null;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);
            _writeTime+=(end-start);

            _lockSub.finishedWrite();
        }
    }

    /**
     * fhrt einen sauberen Shutdown der Datenbank durch
     */
    public String shutdown()
    {
        _lock.requestWrite();

        try
        {
            _con.close();

            _con=null;
            System.gc();

            return null;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            _lock.finishedWrite();
        }
    }

    /**
     * loescht Triplets, zu denen kein Dokument mehr exisitiert
     * loescht Triplets, zu denen kein Server mehr exisitiert
     */
    public String cleanDatabase()
    {
        _calls++;

        long        start=System.currentTimeMillis();
        ResultSet   rs=null;

        try
        {
            rs=_tripletsFiles.executeQuery();

            rs.close();

            while (rs.next())
            {
                int id=rs.getInt(1);

                _deleteTriplet.setInt(1,id);
                _deleteTriplet.executeUpdate();
            }

            rs=_tripletsServers.executeQuery();

            rs.close();

            while (rs.next())
            {
                int id=rs.getInt(1);

                _deleteSubServerTriplets.setInt(1,id);
                _deleteSubServerTriplets.executeUpdate();
            }

            _con.commit();

            return null;
        }
        catch (SQLException ex)
        {
            ex.printStackTrace(Log.getWriter("database","error",0));

            return ex.toString();
        }
        finally
        {
            long    end=System.currentTimeMillis();

            _time+=(end-start);

            try
            {
                if (null!=rs)
                {
                    rs.close();
                }
            }
            catch (SQLException ex)
            {
                ex.printStackTrace(Log.getWriter("database","error",0));
            }
        }
    }

    /**
     * liest eine vernderte Konfiguration ein
     */
    public void update(Node configNode)
    {
        Log.log("database","config",1,"starting reconfiguration");
        _lock.requestWrite();

        try
        {
            synchronized (_syncObj)
            {
                JDataStoreConnection    old=_engine;

                try
                {
                    _engine=new JDataStoreConnection(configNode);
                }
                catch (Exception ex)
                {
                    if (null==_engine)
                    {
                        _engine=old;
                    }
                }

                System.gc();
            }
        }
        catch (Exception ex) {}
        finally
        {
            _lock.finishedWrite();
        }

        Log.log("database","config",1,"finished reconfiguration");
    }

    /**
     * Method declaration
     *
     *
     * @param pw
     *
     * @throws IOException
     *
     * @see
     */
    public void writeStatistic(PrintWriter pw) throws IOException
    {
        pw.println("*** DBConnection");
        pw.println("calls:                   "+_calls);
        pw.println("calls for reading data:  "+_read);
        pw.println("calls for updating data: "+_write);
        pw.println("time:                    "+_time+" ms");
        pw.println("time for reading data:   "+_readTime+" ms");
        pw.println("time for updating data:  "+_writeTime+" ms");
        pw.println();
    }

}



