/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
ldse@hendriklipka.de
*/

package de.hendriklipka.ldse.server.data.document;

import java.util.*;
import java.net.URL;
import java.text.DateFormat;

import de.hendriklipka.ldse.common.Log;
import de.hendriklipka.ldse.common.ResultDescriptor;
import de.hendriklipka.ldse.server.data.URLInfo;

// Beschreibt ein Dokument
public class DocumentDescriptor
{
  protected long _checksum=0;
  protected URL _url=null;
  protected String _urlStr="";
  protected int _spiderID=-1;
  protected int _size=-1;
  protected int _id=-1;
  protected String _title="";
  protected String _excerpt="";
  protected Date _lastChange=null;
  protected TripletList _list=null;
  protected String _type="";
  protected int _tripletCount=0;
  protected int _wordCount=0;
  
  public Vector _words=null;
  
  public DocumentDescriptor(int id, URL url, String urlStr, int spiderID, String title,
                            String excerpt, Date lastChange, long checksum, int size)
  {
    _id=id;
    _url=url;
    _urlStr=urlStr;
    _spiderID=spiderID;
    _title=title;
    _excerpt=excerpt;
    if (null==lastChange)
    {
        lastChange=new Date(); // man weiss ja nie :(
    }
    _lastChange=lastChange;
    _checksum=checksum;
    _size=size;
    if ((null==_url)&&(null!=_urlStr))
    {
      try
      {
        _url=new URL(_urlStr);
      }
      catch (java.net.MalformedURLException ex) {}
    }
    if ((null==_urlStr)&&(null!=_url))
    {
      _urlStr=_url.toString();
    }
  }
  
  public DocumentDescriptor(Properties data)
  {
    String s=data.getProperty("id");
    if (s!=null)
    {
      _id=Integer.parseInt(s);
    }
    s=data.getProperty("url");
    if (s!=null)
    {
      _urlStr=s;
      try
      {
        _url=new URL(s);
      }
      catch (java.net.MalformedURLException ex) {}
    }
    s=data.getProperty("spiderid");
    if (s!=null)
    {
      _spiderID=Integer.parseInt(s);
    }
    s=data.getProperty("title");
    if (s!=null)
    {
      _title=s;
    }
    s=data.getProperty("type");
    if (s!=null)
    {
      _type=s;
    }
    s=data.getProperty("excerpt");
    if (s!=null)
    {
      _excerpt=s;
    }
    s=data.getProperty("last-changed");
    if (s!=null)
    {
      long l=Long.parseLong(s);
      _lastChange=new Date(l);
    }
    s=data.getProperty("checksum");
    if (s!=null)
    {
      _checksum=Long.parseLong(s);
    }
    s=data.getProperty("size");
    if (s!=null)
    {
      _size=Integer.parseInt(s);
    }
    s=data.getProperty("tripletcount");
    if (s!=null)
    {
      _tripletCount=Integer.parseInt(s);
    }
    s=data.getProperty("wordcount");
    if (s!=null)
    {
      _wordCount=Integer.parseInt(s);
    }
  }
  
  public Properties getData()
  {
    Properties prop=new Properties();
    prop.put("id",Integer.toString(_id)); 
    prop.put("url",_url.toString());
    prop.put("spiderid",Integer.toString(_spiderID)); 
    prop.put("title",_title);
    prop.put("excerpt",_excerpt); 
    prop.put("last-changed",Long.toString(_lastChange.getTime()));
    prop.put("checksum",Long.toString(_checksum));
    prop.put("size",Integer.toString(_size));
    prop.put("type",_type);
    prop.put("tripletcount",Integer.toString(_tripletCount));
    prop.put("wordcount",Integer.toString(_wordCount));
    return prop;
  }
  
  public URLInfo getURLInfo()
  {
    URLInfo ui=new URLInfo();
    ui._url=_url;
    ui._urlStr=_urlStr;
    ui._type=_type;
    ui._size=_size;
    return ui;
  }
  
  public long getChecksum()
  {
    return _checksum;
  }
  
  public URL getURL()
  {
    return _url;
  }
  
  public String getUrlStr()
  {
    return _urlStr;
  }
  
  public int getSize()
  {
    return _size;
  }
  
  public int getTripletCount()
  {
    return _tripletCount;
  }
  
  public int getWordCount()
  {
    return _wordCount;
  }
  
  public int getID()
  {
    return _id;
  }
  
  public int getSpiderID()
  {
    return _spiderID;
  }
  
  public String getTitle()
  {
    return _title;
  }
  
  public String getExcerpt()
  {
    return _excerpt;
  }
  
  public Date getLastChange()
  {
    return _lastChange;
  }
  
  public TripletList getList()
  {
    return _list;
  }
  
  public void setList(TripletList list)
  {
    _list=list;
  }
  
  public String getType()
  {
    return _type;
  }
  
  public void setType(String s)
  {
    _type=s;
  }
  
  public void setTripletCount(int num)
  {
    _tripletCount=num;
  }

  public void setWordCount(int num)
  {
    _wordCount=num;
  }
  
  public ResultDescriptor createResultDescriptor()
  {
    ResultDescriptor rd=new ResultDescriptor();
    rd.setURL(_urlStr);
    rd.setSize(_size);
    rd.setLastChange(_lastChange);
    rd.setTitle(_title);
    rd.setExcerpt(_excerpt);
    rd.setType(_type);
    return rd;
  }
}