/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
ldse@hendriklipka.de
*/

package de.hendriklipka.ldse.server.indexer;

import java.util.*;
import java.io.*;

import org.w3c.dom.*;

import de.hendriklipka.ldse.common.Log;
import de.hendriklipka.ldse.server.data.URLInfo;
import de.hendriklipka.ldse.server.data.document.Document;
import de.hendriklipka.ldse.server.data.document.*;
import de.hendriklipka.ldse.server.dataserver.DataClient;

/**
  zerlegt Dokument und indexiert sie
  anschliessend werden sie in die Datenbank geschrieben
*/
public class IndexServer implements de.hendriklipka.ldse.common.Updateable, de.hendriklipka.ldse.common.StatisticWriter
{
  protected DataClient _dataClient=null;
  protected Hashtable _stopwords=new Hashtable();
  
  protected static long _documents=0;
  protected static long _indexedDocuments=0;
  protected static long _words=0;
  protected static long _triplets=0;
  protected static long _time=0;
  protected static long _saveTime=0;
  protected static long _size=0;
  
  public IndexServer(Node configNode, DataClient dataClient)
  {
    _dataClient=dataClient;
    String swfile=null;
    NodeList nl=configNode.getChildNodes();
    for (int i=0;i<nl.getLength();i++)
    {
      Node node=nl.item(i);
      if (node.getNodeType()!=node.ELEMENT_NODE)
        continue;
      String nn=node.getNodeName();
      if (nn.equals("stopwords"))
      {
        swfile=de.hendriklipka.ldse.common.Util.getTextData(node);
      }
    }
    // Liste der Stopwoerter einlesen
    try
    {
      if (null!=swfile)
      {
        StreamTokenizer st=new StreamTokenizer(new FileReader(swfile));
        while (StreamTokenizer.TT_EOF!=st.nextToken())
        {
          String s=(String)st.toString();
          s=s.toLowerCase();
          _stopwords.put(s,s);
        }
      }
    }
    catch (Exception ex)
    {
      ex.printStackTrace(System.err);
    }
  }
  
  public void putURL(URLInfo urlInfo)
  {
    _documents++;
    try
    {
      Document doc=DocumentFactory.getDocument(urlInfo);
      Log.log("indexer","info",1,"putURL for "+urlInfo._urlStr);
      if (null==doc)
      {
        Log.log("indexer","document",0,"no document for "+urlInfo._urlStr);
        return;
      }
      _indexedDocuments++;
      _size+=urlInfo._size;
      // gibt es das Dokument schon?
      if (!_dataClient.existDocument(urlInfo._url))
      {
        // neues Dokument
        long cs1=urlInfo._checksum;
        Log.log("indexer","document",1,"putting new doc");
        Log.log("indexer","document",2,urlInfo._urlStr+": checksum="+cs1);
        if (!doc.init())
        {
          Log.log("indexer","document",0,"can't initialize document for "+urlInfo._urlStr);
          doc.finish();
          return;
        }
        long start=System.currentTimeMillis();
        putDocument(doc);
        long end=System.currentTimeMillis();
        _time+=(end-start);
      }
      else
      {
        // Dok gibt es schon, also holen wir seine Daten
        DocumentDescriptor ds=_dataClient.getDocument(urlInfo._url);
        Log.log("indexer","document",1,"got doc");
        if (null==ds)
        {
          Log.log("indexer","error",1,"no docdesc...");
        }
        if (null==urlInfo)
        {
          Log.log("indexer","error",1,"no urlInfo...");
        }
        long cs1=urlInfo._checksum;
        long cs2=ds.getChecksum();
        Log.log("indexer","document",2,urlInfo._urlStr+": checksum="+cs1+" file checksum="+cs2);
        // Checksumme und Datum gleich??
        if ((cs1!=cs2)||(-1==cs1)||(ds.getLastChange().getTime()!=urlInfo._lastChanged.getTime()))
        {
          // Dokument hat sich geaendert oder Checksum ist unbekannt
          Log.log("indexer","document",1,"putting existing doc");
          // wir muessen es also doch ganz einlesen...
          if (!doc.init())
          {
            Log.log("indexer","document",0,"can't initialize document for "+urlInfo._urlStr);
            return;
          }
          long start=System.currentTimeMillis();
          putDocument(doc);
          long end=System.currentTimeMillis();
          _time+=(end-start);
        }
        else
        {
          // do nothing, document has not changed
          Log.log("indexer","document",1,"unmarking doc");
          _dataClient.unmarkDocument(doc.getDocDesc());
        }
      }
    }
    catch (Throwable ex)
    {
      Log.log("indexer","error",1,"error in putURL");
      ex.printStackTrace(System.err);
      ex.printStackTrace(Log.getWriter("indexer","error",0));
    }
  }
  
  public void putDocument(Document doc)
  {
    Hashtable words=new Hashtable(1000);
    if (null==doc)
    {
      return;
    }
    try
    {
      // indexieren
      int l=0;
      int trCount=0;
      int wordCount=0;
      int redWordCount=0; 
      TripletList tl=new TripletList();
      Log.log("indexer","document",0,"indexing "+doc.getDocDesc().getUrlStr()+" "+doc.getDocDesc().getType());
      while (doc.hasMoreWords())
      {
        String s=doc.nextWord();
        s=s.toLowerCase();
        if (null==s)
          break;
        // dafuer gibt es ohnehin keine Triplets
        if (3>s.length())
            continue;
        // Stopwort?
        if (null!=_stopwords.get(s))
            continue;
        if (s.length()>48)
            s=s.substring(0,48);
        l++;
        // in Triplets zerlegen
        TripletList list=new TripletList(s);
        tl.addList(list);
        trCount+=list.getTripletCount();
        wordCount++;
        
        WordDocument wd=(WordDocument)words.get(s);
        if (null!=wd)
        {
            wd._count++;
        }
        else
        {
            wd=new WordDocument();
            WordData wd2=new WordData();
            wd2._triplets=list;
            wd2._word=s;
            wd._count=1;
            wd._word=wd2;
            words.put(s,wd);
            redWordCount++;
        }
      }
      // das gleiche mit dem Metadaten
      try
      {
        String s=doc.getDescriptiveWords();
        StringTokenizer st=new StringTokenizer(s);
        while (st.hasMoreElements())
        {
          String w=(String)st.nextElement();
          TripletList list=new TripletList(s);
          tl.addList(list,5);
          trCount+=list.getTripletCount();
          wordCount++;
        }
      }
      catch (Exception ex) 
      {
        ex.printStackTrace(Log.getWriter("indexer","error",0));
      }
      _words+=wordCount;
      _triplets+=trCount;
      Log.log("indexer","document",1,"finished indexing "+doc.getDocDesc().getUrlStr()+" with "+wordCount+" words");
      
      // Wort-Daten kopieren
      WordDocument[] wda=new WordDocument[redWordCount];
      redWordCount=0;
      Enumeration e=words.elements();
      while (e.hasMoreElements())
      {
          wda[redWordCount]=(WordDocument)e.nextElement();
          redWordCount++;
      }
      // Dokument abspeichern
      doc.getDocDesc().setList(tl);
      doc.getDocDesc().setTripletCount(trCount);
      doc.getDocDesc().setWordCount(wordCount);
      long start=System.currentTimeMillis();
      _dataClient.putDocument(doc.getDocDesc(),wda);
      long end=System.currentTimeMillis();
      _saveTime+=(end-start);
    }
    catch (Throwable ex)
    {
      ex.printStackTrace(Log.getWriter("indexer","error",0));
    }
    finally
    {
      doc.finish();
    }
  }
  
  public void update(Node configNode)
  {
  }
  
  public void writeStatistic(PrintWriter pw) throws IOException
  {
    pw.println("*** Indexer");
    pw.println("documents:                    "+_documents);
    pw.println("indexed documents:            "+_indexedDocuments);
    pw.println("indexed words:                "+_words);
    pw.println("indexed triplets:             "+_triplets);
    pw.println("size:                         "+_size);
    pw.println("time for indexing and saving: "+_time+" ms");
    pw.println("time for saving:              "+_saveTime+" ms");
    pw.println();
  }
}